from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QCoreApplication)
from qgis.core import (
  QgsRectangle,
  QgsProcessingAlgorithm,
  QgsReferencedRectangle,
  QgsProcessingParameterExtent,
  QgsProcessingParameterCrs,
  QgsProcessingParameterDistance,
  QgsProcessingParameterFeatureSink,
  )
from qgis import processing

from ..hriskapi import HrApi

class fetchbuildingosm(QgsProcessingAlgorithm):
  
  PARAMETERS = {  
    "FETCH_EXTENT": {
      "ui_func": QgsProcessingParameterExtent,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("fetchbuildingosm","Extent for fetching data")
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchbuildingosm","Target CRS (Cartesian coordinates)")
      }
    },
    "BUFFER": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchbuildingosm","Buffer of the fetch area (using Target CRS)"),
        "defaultValue": 0.0,
        "parentParameterName": "TARGET_CRS"
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchbuildingosm","Building")
      }
    }
  }
  
  
  FETCH_TILE_FAMILY = "web_mercator"
  FETCH_BASE_URL = "https://lz4.overpass-api.de/api/interpreter"
  FETCH_TIMEOUT = 30
  FETCH_QUERY_KEY = "building"
  FETCH_QUERY_VALUE = None
  FETCH_ZOOM = None
  FETCH_GEOM_TYPE = "Polygon"
  
  
  def __init__(self):
    super().__init__()
    self.API = HrApi(self)
    
  def initAlgorithm(self, config):    
    (extent, target_crs) = self.API.getExtentAndCrsUsingCanvas()
    self.API.setDefaultValue("FETCH_EXTENT", extent)
    self.API.setDefaultValue("TARGET_CRS", target_crs.authid())
    self.API.initParameters()
  
  def processAlgorithm(self, parameters, context, feedback):
    
    self.API.registerProcessingParameters(parameters, context, feedback)
    self.CURRENT_PROCESS = self.API.parseCurrentProcess()
    # get target x-y CRS, to apply the buffer and determine the fetch area
    target_crs = self.parameterAsCrs(parameters, "TARGET_CRS", context)
    
    # check whether the target CRS is x-y coordinates
    self.API.checkCrsAsCartesian(target_crs)
    
    # get the extent, using the target CRS
    fetch_extent = self.parameterAsExtent(
      parameters, "FETCH_EXTENT", context, 
      self.parameterAsCrs(parameters, "TARGET_CRS", context)
      )
    
    # get the buffer
    buffer = self.parameterAsDouble(parameters, "BUFFER",context)
    
    # get the fetch area, using the extent and buffer
    fetch_area = QgsReferencedRectangle(
      QgsRectangle(
        fetch_extent.xMinimum() - buffer,
        fetch_extent.yMinimum() - buffer,
        fetch_extent.xMaximum() + buffer,
        fetch_extent.yMaximum() + buffer
      ),
      target_crs
    )
    
    fetch_results = processing.run(
      "quickosm:downloadosmdataextentquery",
      {
        "KEY": self.FETCH_QUERY_KEY,
        "VALUE": self.FETCH_QUERY_VALUE,
        "TIMEOUT": self.FETCH_TIMEOUT,
        "SERVER": self.FETCH_BASE_URL,
        "EXTENT": fetch_area,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      # context = context, # not passed, to avoid adding a new layer
      feedback = feedback,
    )["OUTPUT_MULTIPOLYGONS"]
    
    bldg_raw = fetch_results
    
      # transform
    bldg_transformed = processing.run(
    "native:reprojectlayer", 
    {
      "INPUT": bldg_raw,
      "TARGET_CRS": target_crs,
      "OUTPUT": "TEMPORARY_OUTPUT"
    },
    context = context,
    is_child_algorithm = True
    )["OUTPUT"]      
    
    
    
    feedback.pushInfo(self.tr("Snapping and dissolving are applied."))
    # snap geometry
    bldg_snap = processing.run(
      "native:snapgeometries", 
      {
        "INPUT": bldg_transformed,
        "REFERENCE_LAYER": context.getMapLayer(bldg_transformed),
        "TOLERANCE": 0.1,
        "BEHAVIOR": 0,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm = True
    )["OUTPUT"]
      
    # dissolve
    bldg_dissolve = self.API.dissolve(bldg_snap)
      
    
    bldg_final = processing.run(
      "hrisk:initbuilding",{
        "INPUT": bldg_dissolve,
        "TARGET_CRS": target_crs,
        "OVERWRITE": True,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm = True
    )["OUTPUT"]
    
    
    bldg_final_fts = context.getMapLayer(bldg_final)
    new_fields = self.API.newFieldsWithHistory(bldg_final_fts.fields())
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context,
      new_fields, bldg_final_fts.wkbType(), target_crs
    )
    self.API.addFeaturesWithHistoryToSink(
      sink, bldg_final_fts, new_fields,
      current_process=self.CURRENT_PROCESS
    )
    
    return {"OUTPUT": dest_id}
  
  def name(self):
    return self.__class__.__name__

  def displayName(self):
    return self.tr("Building (OSM)")

  def group(self):
    return self.tr('Fetch geometries')

  def groupId(self):
    return 'fetchgeometry'

  def createInstance(self):
    return fetchbuildingosm()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
