from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant, QCoreApplication)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterBoolean,
  QgsFeatureRequest,
  QgsVectorLayer,
  QgsProcessingParameterCrs,
  QgsField
  )


from qgis import processing

from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrFields, HrPostProcessor

class initreceiver(QgsProcessingAlgorithm):
  PARAMETERS = {                  
    "INPUT": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("initreceiver","Point layer"),
        "types": [QgsProcessing.TypeVectorPoint],
        "optional": True
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("initreceiver","Target CRS (Cartesian coordinates)")
      }
    },
    "OVERWRITE": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args":{
        "description" : QT_TRANSLATE_NOOP("initreceiver","Overwrite existing fields?"),
        "defaultValue": True
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("initreceiver","Elevation points" )
      }
    }
  }
  
  FIELDS = {}
  
  def __init__(self):
    super().__init__()
    self.API = HrApi(self)    
    self.FIELDS["COMMON"] = HrFields.fromQgsFieldList([
      QgsField("PK", QVariant.Int, "int"),
      QgsField("HISTORY", QVariant.String, "string")      
    ]).setComments("note: common fields; ")
      
  def initAlgorithm(self, config):
    (extent, target_crs) = self.API.getExtentAndCrsUsingCanvas()
    self.API.setDefaultValue("TARGET_CRS", target_crs.authid())
    self.API.initParameters()
    
  def processAlgorithm(self, parameters, context, feedback):    
    self.API.registerProcessingParameters(parameters, context, feedback)    
    target_crs = self.parameterAsCrs(parameters, "TARGET_CRS", context)
    # check whether the target CRS is x-y coordinates
    self.API.checkCrsAsCartesian(target_crs)

    try:
      input_lyr_proj = processing.run(
        "native:reprojectlayer", 
        {
          "INPUT": self.parameterAsSource(parameters, "INPUT", context).materialize(QgsFeatureRequest(), feedback),
          "TARGET_CRS": target_crs,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]      
      input_lyr_fixed = context.getMapLayer(input_lyr_proj)
            
    except:
      input_lyr_proj = QgsVectorLayer(f"Polygon?crs={target_crs.authid()}", "temp", "memory")
      input_lyr_fixed = input_lyr_proj

    try:
      self.FIELDS["EXISTING"] = HrFields.fromQgsFields(
        input_lyr_fixed.fields()
      ).setComments("note: existing fields; ", append = False)
    except:
      feedback.reportError(self.tr("The fields of input layer is invalid. The names may be duplicated (case-insensitive)."))
      raise Exception
    
    self.FIELDS["USE"] = self.FIELDS["COMMON"].concat(
      self.FIELDS["EXISTING"],
      overwrite = not self.parameterAsBoolean(parameters, "OVERWRITE", context)
    )
    
    # feedback the results
    feedback.pushInfo(self.tr("The following fields are initialized: "))
    feedback.pushInfo(str(self.FIELDS["USE"]))    
    
    
    # define the feature sink
    new_fields = self.API.newFieldsWithHistory(self.FIELDS["USE"])
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context, 
      new_fields, 
      input_lyr_fixed.wkbType(), 
      input_lyr_fixed.sourceCrs()
    )
    
    self.CURRENT_PROCESS = self.API.parseCurrentProcess()
    n = input_lyr_fixed.featureCount()
    additional_attributes = {
      "PK": list(range(1, n+1))
    }
        
    self.API.addFeaturesWithHistoryToSink(
      sink = sink, vector_layer = input_lyr_fixed, fields = new_fields, 
      current_process = self.CURRENT_PROCESS,
      additional_attributes = additional_attributes
    )
    PostProcessors[dest_id] = HrPostProcessor(history = [self.CURRENT_PROCESS])
    
    return {"OUTPUT": dest_id}
  
  def name(self):
    return self.__class__.__name__
  
  def displayName(self):
    return self.tr("Receiver")

  def group(self):
    return self.tr("Initialize features")

  def groupId(self):
    return "initfeature"

  def createInstance(self):
    return initreceiver()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
