from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant, QCoreApplication)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterNumber,
  QgsProcessingParameterString,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterField,
  QgsFeatureRequest,
  )

from qgis import processing
from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrPostProcessor

import os

class isosurface(QgsProcessingAlgorithm):
  PARAMETERS = { 
    "LEVEL_RESULT": {
      "crs_referrence": True, # this parameter is used as CRS referrence
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("isosurface","Sound level layer"),
        "types": [QgsProcessing.TypeVectorPoint]
      }
    },
    "LEVEL_RID": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("isosurface","Receiver ID Field of the Sound level layer"),
        "parentLayerParameterName": "LEVEL_RESULT",
        "defaultValue": "IDRECEIVER"
      }
    },
    "LEVEL_LID": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("isosurface","Sound-level Field of the Sound level layer"),
        "parentLayerParameterName": "LEVEL_RESULT",
        "defaultValue": "LAEQ"
      }
    },
    "TRIANGLES": { # Note: the name "TRIANGLES" should not be changed
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("isosurface","Triangle layer"),
        "types": [QgsProcessing.TypeVectorPolygon]
      },
      "n_mdl": "triangleGeomPath",
      "save_layer_get_path": True
    },
    "ISO_CLASS": {
      "ui_func": QgsProcessingParameterString,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("isosurface","Separation of sound levels for isosurfaces (e.g. 35.0,40.0)"),
        "defaultValue": "35.0,40.0,45.0,50.0,55.0,60.0,65.0,70.0,75.0,80.0,200.0",
        "multiLine": False
      },
      "n_mdl": "isoClass"
    },    
    "SMOOTH_COEF": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("isosurface","Smoothing parameter (Bezier curve coefficient)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 0.0, "defaultValue": 1.0, "maxValue": 2.0
      },
      "n_mdl": "smoothCoefficient"
    },
    "OPACITY": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args":{
        "optional": True,
        "description": QT_TRANSLATE_NOOP("isosurface","Opacity of the fill (0: transparent - 1: opaque)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 0.0, "defaultValue": None, "maxValue": 1.0,
        "defaultValue": 1.0
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("isosurface","Isosurface")
      }
    }
  }
        
  def __init__(self):
    super().__init__()
    self.API = HrApi(self)  
    self.BLDG_LEVEL_ARGS = {
      "BUILDING_BID": "PK", # primary key of the buildings
      "RECEIVER_BID": "BUILD_PK", # primary key of the buildings, written in receiver layer
      "RECEIVER_RID": "PK", # primary key of the receivers
      "LEVEL_RID": "IDRECEIVER", #primary key of the receivers, written in level layer    
      "LEVEL_ASSIGN": ["LAEQ"], # which attribute(s) to be assined,
      "OVERWRITE": True
    }
  
  def initAlgorithm(self, config):
    self.API.initParameters()

  def processAlgorithm(self, parameters, context, feedback):
    
    self.API.registerProcessingParameters(parameters, context, feedback) 
    self.CURRENT_PROCESS = self.API.parseCurrentProcess(with_nm=True)
    target_crs = self.API.parseCrs()
    
    triangles = self.parameterAsSource(parameters, "TRIANGLES", context).materialize(QgsFeatureRequest(), feedback)
    assert "PK_1" in triangles.fields().names(), "PK_1 not found in the triangle layer"    
    assert "PK_2" in triangles.fields().names(), "PK_2 not found in the triangle layer"    
    assert "PK_3" in triangles.fields().names(), "PK_3 not found in the triangle layer"    
    assert "CELL_ID" in triangles.fields().names(), "CELL_ID not found in the triangle layer"    
    
    self.NOISEMODELLING = self.API.parseNoiseModellingArgs(
      {
        "GROOVY_SCRIPT": os.path.join("%grvhome%", "isosurface.groovy"),
        "ISOSURFACE": os.path.join("%nmtmp%", "CONTOURING_NOISE_MAP.geojson"),
        "ISOSURFACE_REV": os.path.join("%nmtmp%", "CONTOURING_NOISE_MAP_REV.geojson")
      },
      target_crs = target_crs,
      ui_parameters = self.PARAMETERS,
      wps_args={
        "resultGeomPath": os.path.join(self.API.createTempDir(), "LEVEL_TMP.geojson"),
      }
    )
    
    spl_input = self.parameterAsSource(parameters, "LEVEL_RESULT", context).materialize(QgsFeatureRequest(), feedback)
    pk_field = self.parameterAsFields(parameters, "LEVEL_RID", context)[0]
    level_field = self.parameterAsFields(parameters, "LEVEL_LID", context)[0]
    
    spl_ext = processing.run(
      "native:retainfields",
      {
        "INPUT": spl_input,
        "FIELDS": [pk_field, level_field],
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm = True
    )["OUTPUT"]
    
    if pk_field != "PK":
      spl_pk = processing.run(
        "native:renametablefield",
        {
          "INPUT": spl_ext,
          "FIELD": pk_field,
          "NEW_NAME": "PK",
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]
    else:
      spl_pk = spl_ext
    
    if level_field != "LAEQ":
      spl_laeq = processing.run(
        "native:renametablefield",
        {
          "INPUT": spl_pk,
          "FIELD": level_field,
          "NEW_NAME": "LAEQ",
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]
    else:
      spl_laeq = spl_pk
    
    self.API.saveVectorLayer(
      context.getMapLayer(spl_laeq), 
      self.NOISEMODELLING["WPS_ARGS"]["resultGeomPath"]
    )
                
    feedback.pushCommandInfo("CMD FOR WPS_SCRIPTS: " + self.NOISEMODELLING["CMD"])
    self.API.execNoiseModellingCmd(
      cmd      = self.NOISEMODELLING["CMD"],
      temp_dir = self.NOISEMODELLING["TEMP_DIR"]
    )
    feedback.setProgress(100)
    
    # add level field
    iso_breaks = [-999.0] + [float(x) for x in self.parameterAsString(parameters, "ISO_CLASS", context).split(",")]
    iso_mean = []
    for i in range(1, len(iso_breaks)):
      iso_mean.append((iso_breaks[i] + iso_breaks[i - 1]) / 2)
    
    isosurface_layer = processing.run(
      "native:fieldcalculator",
      {
        "INPUT": self.NOISEMODELLING["ISOSURFACE"],
        "FIELD_NAME": "ISOLVL_MEAN",
        "FIELD_TYPE": 0,
        "FIELD_LENGTH": 10,
        "FIELD_PRECISION": 3,
        "FORMULA": "array(" + ",".join([str(x) for x in iso_mean]) + ")[\"ISOLVL\"]",
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      is_child_algorithm = True,
      context = context
    )["OUTPUT"]
    isosurface_layer = context.getMapLayer(isosurface_layer)
      
    # import the result
    new_fields = self.API.newFieldsWithHistory(isosurface_layer.fields())
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context, 
      new_fields, isosurface_layer.wkbType(), target_crs
    )
    self.API.addFeaturesWithHistoryToSink(
      sink, isosurface_layer, new_fields, self.CURRENT_PROCESS
    )
    
    PostProcessors[dest_id] = HrPostProcessor(
      history = [self.CURRENT_PROCESS],
      coloring = "graduated_symbol", 
      attribute = "ISOLVL_MEAN", 
      opacity = parameters.get("OPACITY"),
      theme = "Default"
      )
    
    self.API.registerPostProcessAlgorithm(context)
    
    return {"OUTPUT": dest_id}          

  def name(self):
    return self.__class__.__name__
  
  def displayName(self):
    return self.tr("Isosurface")

  def group(self):
    return self.tr('Predict sound level')

  def groupId(self):
    return 'soundlevel'

  def createInstance(self):
    return isosurface()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
