from qgis.PyQt.QtCore import (QCoreApplication, QT_TRANSLATE_NOOP)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterNumber,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterExtent,
  QgsVectorLayer
  )


from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrPostProcessor
import os

class receiverdelaunaygrid(QgsProcessingAlgorithm):
  PARAMETERS = { 
    "BUILDING": {
      "crs_referrence": True, # this parameter is used as CRS referrence
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Building layer"),
        "types": [QgsProcessing.TypeVectorPolygon]
      },
      "n_mdl": "buildingGeomPath",
      "save_layer_get_path": True
    },    
    "SOURCE": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Source layer"),
        "types": [QgsProcessing.TypeVectorPoint,QgsProcessing.TypeVectorLine,QgsProcessing.TypeVectorPolygon]
      },
      "n_mdl": "sourceGeomPath",
      "save_layer_get_path": True
    },    
    "FENCE_EXTENT": {
      "ui_func": QgsProcessingParameterExtent,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Calculation extent"),
        "defaultValue": None,
        "optional": True,
      }
    },
    "MAX_CELL_DIST": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Maximum distance used to split the domain into sub-domains (m)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 100.0, "defaultValue": 600.0, "maxValue": 2000.0
      },
      "n_mdl": "maxCellDist"
    },
    "ROAD_WIDTH": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Road width (m), where no receivers will be set closer than it"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 1.0, "defaultValue": 2.0, "maxValue": 20.0
      },
      "n_mdl": "roadWidth"
    },
    "MAX_AREA": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Maximum trianglar area (m2)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 10.0, "defaultValue": 500.0, "maxValue": 10000.0
      },
      "n_mdl": "maxArea"
    },
    "HEIGHT": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Height of receivers (m)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 0.01, "defaultValue": 4.0, "maxValue": 100.0
      },
      "n_mdl": "height"
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Receivers of Delaunay" )
      }
    },
    "TRIANGLE": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverdelaunaygrid","Triangles of Delaunay" )
      }
    }

  }
  
  def __init__(self) -> None:
    super().__init__()
    self.API = HrApi(self)
    
    
  def initAlgorithm(self, config):
    self.API.initParameters()

  def processAlgorithm(self, parameters, context, feedback):  
    
    self.API.registerProcessingParameters(parameters, context, feedback)
    self.CURRENT_PROCESS = self.API.parseCurrentProcess(with_nm=True)
    target_crs = self.API.parseCrs()
    
    fence_extent = self.parameterAsString(parameters, "FENCE_EXTENT", context)
    if len(fence_extent) > 0:
      wps_args = {
        "fence": '"' + self.API.fenceExtentAsWkt(context, fence_extent, target_crs) + '"'
      }
    else:
      wps_args = {}
    
    self.NOISEMODELLING = self.API.parseNoiseModellingArgs(
      {
        "GROOVY_SCRIPT": os.path.join("%grvhome%", "receiverdelaunaygrid.groovy"),
        "RECEIVER": os.path.join("%nmtmp%", "RECEIVERS.geojson"),
        "TRIANGLE": os.path.join("%nmtmp%", "TRIANGLES.geojson")
      },
      target_crs = target_crs,
      wps_args = wps_args,
      ui_parameters = self.PARAMETERS
    )
    
    feedback.pushCommandInfo("CMD FOR WPS_SCRIPTS: " + self.NOISEMODELLING["CMD"])
    self.API.execNoiseModellingCmd(
      cmd      = self.NOISEMODELLING["CMD"],
      temp_dir = self.NOISEMODELLING["TEMP_DIR"]
    )
    feedback.setProgress(100)
    
    receiver_result = QgsVectorLayer(self.NOISEMODELLING["RECEIVER"])
    triangle_result = QgsVectorLayer(self.NOISEMODELLING["TRIANGLE"])
    
    rcv_fields = self.API.newFieldsWithHistory(receiver_result.fields())
    tri_fields = self.API.newFieldsWithHistory(triangle_result.fields())
    
    (sink_rcv, dest_id_rcv) = self.parameterAsSink(
      parameters, "OUTPUT", context, 
      rcv_fields, receiver_result.wkbType(), target_crs
    )
    (sink_tri, dest_id_tri) = self.parameterAsSink(
      parameters, "TRIANGLE", context, 
      tri_fields, triangle_result.wkbType(), target_crs
    )

    self.API.addFeaturesWithHistoryToSink(
      sink_rcv, receiver_result, rcv_fields, self.CURRENT_PROCESS
    )

    self.API.addFeaturesWithHistoryToSink(
      sink_tri, triangle_result, tri_fields, self.CURRENT_PROCESS
    )
    
    PostProcessors[dest_id_rcv] = HrPostProcessor(history = [self.CURRENT_PROCESS])
    PostProcessors[dest_id_tri] = HrPostProcessor(history = [self.CURRENT_PROCESS])
    
    return {"OUTPUT": dest_id_rcv, "TRIANGLE": dest_id_tri}
  
  def name(self):
    return self.__class__.__name__

  def displayName(self):
    return self.tr("Delaunay grid")

  def group(self):
    return self.tr('Set receivers')

  def groupId(self):
    return 'receiver'

  def createInstance(self):
    return receiverdelaunaygrid()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
