from qgis.PyQt.QtCore import (QCoreApplication, QT_TRANSLATE_NOOP, QVariant)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterNumber,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterExtent,
  QgsVectorLayer
  )


from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrPostProcessor

import os

class receiverregulargrid(QgsProcessingAlgorithm):
  PARAMETERS = { 
    "FENCE_EXTENT": {
      "ui_func": QgsProcessingParameterExtent,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("receiverfacade","Calculation extent"),
        "defaultValue": None
      }
    },
    "BUILDING": {
      "crs_referrence": True, # this parameter is used as CRS referrence
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("receiverregulargrid","Building layer"),
        "types": [QgsProcessing.TypeVectorPolygon],
        "optional": True
      },
      "n_mdl": "buildingGeomPath",
      "save_layer_get_path": True
    },    
    "SOURCE": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("receiverregulargrid","Source layer"),
        "types": [QgsProcessing.TypeVectorPoint,QgsProcessing.TypeVectorLine,QgsProcessing.TypeVectorPolygon],
        "optional": True,
      },
      "n_mdl": "sourceGeomPath",
      "save_layer_get_path": True
    },    
    
    "DELTA": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverregulargrid","Distance between receivers (m)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 1.0, "defaultValue": 10.0, "maxValue": 100.0
      },
      "n_mdl": "delta"
    },
    "HEIGHT": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverregulargrid","Height of receivers (m)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 0.01, "defaultValue": 4.0, "maxValue": 100.0
      },
      "n_mdl": "height"
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("receiverregulargrid","Receivers at regular grid" )
      }     
    }
  }
  
  def __init__(self) -> None:
    super().__init__()
    self.API = HrApi(self)
    
  def initAlgorithm(self, config):
    self.API.initParameters()
  

  def processAlgorithm(self, parameters, context, feedback):    
    
    self.API.registerProcessingParameters(parameters, context, feedback)
    self.CURRENT_PROCESS = self.API.parseCurrentProcess(with_nm=True)
    target_crs = self.API.parseCrs()
    
    fence_extent = self.parameterAsString(parameters, "FENCE_EXTENT", context)
    if len(fence_extent) > 0:
      fence_layer = self.API.fenceExtentAsLayer(context, fence_extent, target_crs)
      fence_path = os.path.join(self.API.createTempDir(), "FENCE.geojson")
      self.API.saveVectorLayer(context.getMapLayer(fence_layer), fence_path)
      wps_args = {"fenceGeomPath": '"' + fence_path + '"'}
    else:
      wps_args = {}
      
    self.NOISEMODELLING = self.API.parseNoiseModellingArgs(
      {
        "GROOVY_SCRIPT": os.path.join("%grvhome%", "receiverregulargrid.groovy"),
        "RECEIVER": os.path.join("%nmtmp%", "RECEIVERS.geojson")
      },
      target_crs = target_crs,
      wps_args = wps_args,
      ui_parameters = self.PARAMETERS
    )
    
    feedback.pushCommandInfo("CMD FOR WPS_SCRIPTS: " + self.NOISEMODELLING["CMD"])
    self.API.execNoiseModellingCmd(
      cmd      = self.NOISEMODELLING["CMD"],
      temp_dir = self.NOISEMODELLING["TEMP_DIR"]
    )
    feedback.setProgress(100)
    
    receiver_result = QgsVectorLayer(self.NOISEMODELLING["RECEIVER"])
    rcv_fields = self.API.newFieldsWithHistory(receiver_result.fields())
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context, 
      rcv_fields, receiver_result.wkbType(), target_crs
    )
    
    self.API.addFeaturesWithHistoryToSink(
      sink, receiver_result, rcv_fields, self.CURRENT_PROCESS
    )
    PostProcessors[dest_id] = HrPostProcessor(history = [self.CURRENT_PROCESS])
    
    return {"OUTPUT": dest_id}
          

  def name(self):
    return self.__class__.__name__
  
  def displayName(self):
    return self.tr("Regular grid")

  def group(self):
    return self.tr('Set receivers')

  def groupId(self):
    return 'receiver'

  def createInstance(self):
    return receiverregulargrid()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
