from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant, QCoreApplication)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterBoolean,
  QgsProcessingParameterString,
  QgsFeatureRequest,
  QgsProcessingParameterField
  )
from qgis import processing
from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrPostProcessor

class summarizerisk(QgsProcessingAlgorithm):
  PARAMETERS = {
    "BUILDING": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("summarizerisk","Building layer, with population and level"),
        "types": [QgsProcessing.TypeVectorPolygon]
      }
    },
    "GROUPING": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("summarizerisk","Do the results need to be grouped?"),
        "defaultValue": True
      }
    },
    "GROUPING_FIELD": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("summarizerisk","Field to group the data"),
        "parentLayerParameterName": "BUILDING",
        "defaultValue": "LDEN_LAEQ_Maximum"
      }
    },
    "GROUP_BREAKPOINTS": {
      "ui_func": QgsProcessingParameterString,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("summarizerisk","Break-points of the grouping"),
        "defaultValue": "50, 55, 60, 65, 70"
      }
    },
    "LDEN_FIELD": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("summarizerisk","Lden field"),
        "parentLayerParameterName": "BUILDING",
        "defaultValue": "LDEN_LAEQ_Maximum",
        "optional": True
      }
    },
    "LNIGHT_FIELD": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("summarizerisk","Lnight field"),
        "parentLayerParameterName": "BUILDING",
        "defaultValue": "LNIGHT_LAEQ_Maximum",
        "optional": True
      }
    },
    "POP_FIELD": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("summarizerisk","Population field"),
        "parentLayerParameterName": "BUILDING",
        "defaultValue": "pop",
        "optional": True
      }
    },
    "RECALCULATE_RISK": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("summarizerisk","Re-calculate the health risks?"),
        "defaultValue": False
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("summarizerisk","Building group with health risks")
      }
    }
  }
  
  AGGREGATES = {
    "group":       {"input": None,            "aggregate": "group", "type": 10}, # special aggregation for categorized variable
    "nBldg":       {"input": '"fid"',         "aggregate": "count", "type": 2},
    "areaBldg":    {"input": '"areaBldg"',    "aggregate": "sum",   "type": 6},
    "nPop":        {"input": '"pop"',         "aggregate": "sum",   "type": 6},
    "nHA":         {"input": '"nHA"',         "aggregate": "sum",   "type": 6},
    "nHSD":        {"input": '"nHSD"',        "aggregate": "sum",   "type": 6},
    "nPatientIHD": {"input": '"nPatientIHD"', "aggregate": "sum",   "type": 6},
    "nIncIHD":     {"input": '"nIncIHD"',     "aggregate": "sum",   "type": 6},
    "nDeathIHD":   {"input": '"nDeathIHD"',   "aggregate": "sum",   "type": 6}
  }
  
  def __init__(self):
    super().__init__()
    self.API = HrApi(self)
  
  def initAlgorithm(self, config):
    self.API.initParameters()
  
  # convert the data type number to string
  def getDtypeStr(self, dtype_num, long = False):
    if dtype_num == QVariant.Double:
      return "double precision" if long else "double"
    elif dtype_num == QVariant.Int:
      return "integer" if long else "int"
    elif dtype_num == QVariant.String:
      return "text" if long else "text"
    else:
      return "not num"
    
  # compute the expression for aggregation
  def cmptAggExpr(self, feature, group_expr, feedback):  
    aggregates = []
    for k, v in self.AGGREGATES.items():
      agg_dict = {"name": k}
      agg_dict.update(v)
      
      # check if the input field exists
      if agg_dict["input"].replace("\"","") not in feature.fields().names():
        continue
        
      # data types of input and output
      input_dtype = self.getDtypeStr(feature.fields().field(feature.fields().indexFromName(agg_dict["input"].replace("\"",""))).type())
      output_dtype = agg_dict["type"]
      
      # check if the aggregation method is the special method: group
      if agg_dict["aggregate"] == "group":
        if group_expr == "NULL":
          feedback.pushInfo("No grouping is applied")
          continue
        else:
          feedback.pushInfo(f"Grouping using {agg_dict['input']} is applied")
          agg_dict["input"] = group_expr.replace(f"\"{agg_dict['input']}\"", f"median(\"{agg_dict['input']}\")")
          agg_dict["aggregate"] = "first_value"
      elif agg_dict["aggregate"] == "sum":
        if input_dtype == "double":
          if output_dtype == "int":
            feedback.pushWarning(f"{agg_dict['name']} int = sum ({agg_dict['input']}) double")
          else:
            feedback.pushInfo(f"{agg_dict['name']} = sum ({agg_dict['input']})")            
        elif input_dtype == "int":
          if output_dtype == "double":
            feedback.pushWarning(f"{agg_dict['name']} double = sum ({agg_dict['input']}) int")
          else:
            feedback.pushInfo(f"{agg_dict['name']} = sum ({agg_dict['input']})")            
        elif input_dtype == "text" or input_dtype == "not num":
          feedback.pushWarning(f"{agg_dict['input']} is not a numeric field and skipped calculating {agg_dict['name']}")
          continue
      
      # set the residual parameters for the aggregation and add to the list
      agg_dict["delimiter"] = "."
      agg_dict["length"] = 0
      agg_dict["precision"] = 0
      agg_dict["sub_type"] = 0
      agg_dict["type_name"] = self.getDtypeStr(agg_dict["type"], long = True)
      aggregates.append(agg_dict) 

    return aggregates
  
  # compute the expression for grouping
  def cmptGroupExpr(self, var_str, grouping_str):
    exp_output = f"CASE WHEN \"{var_str}\" is NULL THEN 'NULL'"
    grp_prev = ""
    for grp in grouping_str.split(","):
      try:
        grp = float(grp)
      except:
        raise Exception(self.tr("Invalid value for grouping"))
      if grp_prev == "":
        exp_output += f" WHEN \"{var_str}\" < {grp} THEN '< {grp}'"
      else:
        exp_output += f" WHEN \"{var_str}\" < {grp} THEN '{grp_prev} - {grp}'"
      grp_prev = grp
    return exp_output + f" ELSE '>= {grp_prev}' END"

  def processAlgorithm(self, parameters, context, feedback):    
    
    self.API.registerProcessingParameters(parameters, context, feedback)
    self.CURRENT_PROCESS = self.API.parseCurrentProcess()
    
    # recalculate the health risks, if necessary
    if self.parameterAsBoolean(parameters, "RECALCULATE_RISK", context):
      bldg_risk = processing.run(
        "algrisk:estimateriskofbuilding",
        {
          "BUILDING": self.parameterAsSource(parameters, "BUILDING", context),
          "LDEN": self.parameterAsFields(parameters, "LDEN_FIELD", context)[0],
          "LNIGHT": self.parameterAsFields(parameters, "LNIGHT_FIELD", context)[0],
          "POP": self.parameterAsFields(parameters, "POP_FIELD", context)[0],
          "OUTPUT": "memory:"
        },
        context=context,
        feedback=feedback,
        is_child_algorithm=True
      )["OUTPUT"]
      bldg_risk = context.getMapLayer(bldg_risk)
    else:
      bldg_risk = self.parameterAsSource(parameters, "BUILDING", context).materialize(QgsFeatureRequest(), feedback)
    
    # check if the results need to be grouped
    if self.parameterAsBoolean(parameters, "GROUPING", context):
      grouping_field = self.parameterAsString(parameters, "GROUPING_FIELD", context)
      grouping_expr = self.cmptGroupExpr(grouping_field, self.parameterAsString(parameters, "GROUP_BREAKPOINTS", context))
      self.AGGREGATES["group"]["input"] = grouping_field
      
      # compute the expression for aggregation
      try:
        aggregates = self.cmptAggExpr(bldg_risk, grouping_expr, feedback=feedback)
        if len(aggregates) == 0:
          raise Exception(self.tr("No valid field for aggregation"))
      except Exception as e:
        feedback.reportError(str(e))
        raise e
    else:
      grouping_expr = "NULL"
      aggregates = self.cmptAggExpr(bldg_risk, grouping_expr, feedback=feedback)
    
    # aggregate the data
    bldg_grouped = processing.run(
      "native:aggregate", 
      {
        "INPUT": bldg_risk,
        "GROUP_BY": grouping_expr,
        "AGGREGATES": aggregates,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context=context,
      feedback=feedback,
      is_child_algorithm=True
    )["OUTPUT"]
    
    bldg_grouped = context.getMapLayer(bldg_grouped)
    new_fields = self.API.newFieldsWithHistory(bldg_grouped.fields())
    
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context,
      new_fields, bldg_grouped.wkbType(), bldg_grouped.sourceCrs()
    )
    
    self.API.addFeaturesWithHistoryToSink(
      sink, bldg_grouped, new_fields, self.CURRENT_PROCESS
    )      
    
    PostProcessors[dest_id] = HrPostProcessor(history = [self.CURRENT_PROCESS])
    self.API.registerPostProcessAlgorithm(context)
          
    return {"OUTPUT": dest_id}
  

  def name(self):
    return self.__class__.__name__

  def displayName(self):
    return self.tr("Summarize Risks")

  def group(self):
    return self.tr("Evaluate health risk")

  def groupId(self):
    return "healthrisk"
  
  def createInstance(self):
    return summarizerisk()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
